from contextlib import asynccontextmanager
from typing import Optional

import aiohttp
from serde.json import from_json

from aiochris.client.authed import AuthenticatedClient
from aiochris.errors import raise_for_status
from aiochris.models.collection_links import CollectionLinks
from aiochris.models.data import UserData
from aiochris.types import ChrisURL, Username, Password


class ChrisClient(AuthenticatedClient[CollectionLinks, "ChrisClient"]):
    """
    A normal user *ChRIS* client, who may upload files and create plugin instances.
    """

    @classmethod
    async def create_user(
        cls,
        url: ChrisURL | str,
        username: Username | str,
        password: Password | str,
        email: str,
        session: Optional[aiohttp.ClientSession] = None,
    ) -> UserData:
        payload = {
            "template": {
                "data": [
                    {"name": "email", "value": email},
                    {"name": "username", "value": username},
                    {"name": "password", "value": password},
                ]
            }
        }
        headers = {
            "Content-Type": "application/vnd.collection+json",
            "Accept": "application/json",
        }
        async with _optional_session(session) as session:
            res = await session.post(url + "users/", json=payload, headers=headers)
            await raise_for_status(res)
            return from_json(UserData, await res.text())


@asynccontextmanager
async def _optional_session(session: Optional[aiohttp.ClientSession]):
    if session is not None:
        yield session
        return
    async with aiohttp.ClientSession() as session:
        yield session



import abc
from typing import AsyncContextManager, Generic, Optional, Callable, TypeVar

import aiohttp
from serde import from_dict

from aiochris import Search
from aiochris.link.collection_client import L, CollectionJsonApiClient
from aiochris.models.public import PublicPlugin
from aiochris.errors import raise_for_status

CSelf = TypeVar(
    "CSelf", bound="BaseChrisClient"
)  # can't wait for `Self` in Python 3.11!


class BaseChrisClient(
    Generic[L, CSelf],
    CollectionJsonApiClient[L],
    AsyncContextManager[CSelf],
    abc.ABC,
):
    """
    Provides the implementation for most of the read-only GET resources of ChRIS
    and functions related to the client object's own usage.
    """

    @classmethod
    async def new(
        cls,
        url: str,
        max_search_requests: int = 100,
        connector: Optional[aiohttp.BaseConnector] = None,
        connector_owner: bool = True,
        session_modifier: Optional[Callable[[aiohttp.ClientSession], None]] = None,
    ) -> CSelf:
        """
        A constructor which creates the session for the `BaseChrisClient`
        and makes an initial request to populate `collection_links`.

        Parameters
        ----------
        url
            ChRIS backend url, e.g. "https://cube.chrisproject.org/api/v1/"
        max_search_requests
            Maximum number of HTTP requests to make while retrieving items from a
            paginated endpoint before raising `aiochris.util.search.TooMuchPaginationError`.
            Use `max_search_requests=-1` to allow for "infinite" pagination
            (well, you're still limited by Python's stack).
        connector
            [`aiohttp.BaseConnector`](https://docs.aiohttp.org/en/v3.8.3/client_advanced.html#connectors) to use.
            If creating multiple client objects in the same program,
            reusing connectors between them is more efficient.
        connector_owner
            If `True`, this client will close its `aiohttp.BaseConnector`
        session_modifier
            Called to mutate the created `aiohttp.ClientSession` for the object.
            If the client requires authentication, define `session_modifier`
            to add authentication headers to the session.
        """
        if not url.endswith("/api/v1/"):
            raise ValueError("url must end with /api/v1/")
        accept_json = {
            "Accept": "application/json",
            # 'Content-Type': 'application/vnd.collection+json',
        }
        # TODO maybe we want to wrap the session:
        # - status == 4XX --> print response text
        # - content-type: application/vnd.collection+json
        session = aiohttp.ClientSession(
            headers=accept_json,
            raise_for_status=False,
            connector=connector,
            connector_owner=connector_owner,
        )
        if session_modifier is not None:
            session_modifier(session)
        try:
            async with session.get(url) as res:
                await raise_for_status(res)
                body = await res.json()
        except Exception:
            await session.close()
            raise
        links = from_dict(cls._collection_type(), body["collection_links"])
        return cls(
            url=url,
            s=session,
            collection_links=links,
            max_search_requests=max_search_requests,
        )

    async def __aenter__(self) -> CSelf:
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close()

    async def close(self):
        """
        Close the HTTP session used by this client.
        """
        await self.s.close()

    @abc.abstractmethod
    def search_plugins(self, **query) -> Search[PublicPlugin]:
        """
        Search for plugins.
        """
        ...



import abc
import os
from pathlib import Path
from typing import Optional, Generic, Callable, Sequence

import aiohttp
from async_property import async_cached_property

from aiochris.client.base import BaseChrisClient
from aiochris.client.base import L, CSelf
from aiochris.link import http
from aiochris.link.linked import deserialize_res
from aiochris.models.logged_in import Plugin, File, User, PluginInstance, Feed, PACSFile
from aiochris.models.public import ComputeResource
from aiochris.types import ChrisURL, Username, Password
from aiochris.errors import IncorrectLoginError, raise_for_status
from aiochris.util.search import Search, acollect
from aiochris.client.from_chrs import ChrsLogins


class AuthenticatedClient(BaseChrisClient[L, CSelf], Generic[L, CSelf], abc.ABC):
    """
    An authenticated ChRIS client.
    """

    @classmethod
    async def from_login(
        cls,
        url: str | ChrisURL,
        username: str | Username,
        password: str | Password,
        max_search_requests: int = 100,
        connector: Optional[aiohttp.TCPConnector] = None,
        connector_owner: bool = True,
    ) -> CSelf:
        """
        Get authentication token using username and password, then construct the client.

        See `aiochris.client.base.BaseChrisClient.new` for parameter documentation.
        """
        async with aiohttp.ClientSession(
            connector=connector, connector_owner=False
        ) as session:
            try:
                c = await cls.__from_login_with(
                    url=url,
                    username=username,
                    password=password,
                    max_search_requests=max_search_requests,
                    session=session,
                    connector_owner=connector_owner,
                )
            except BaseException as e:
                if connector is None:
                    await session.connector.close()
                raise e
        return c

    @classmethod
    async def __from_login_with(
        cls,
        url: str | ChrisURL,
        username: Username,
        password: Password,
        max_search_requests: int,
        session: aiohttp.ClientSession,
        connector_owner: bool,
    ) -> CSelf:
        """
        Get authentication token using the given session, and then construct the client.
        """
        payload = {"username": username, "password": password}
        login = await session.post(url + "auth-token/", json=payload)
        if login.status == 400:
            raise IncorrectLoginError(await login.text())
        await raise_for_status(login)
        data = await login.json()
        return await cls.from_token(
            url=url,
            token=data["token"],
            max_search_requests=max_search_requests,
            connector=session.connector,
            connector_owner=connector_owner,
        )

    @classmethod
    async def from_token(
        cls,
        url: str | ChrisURL,
        token: str,
        max_search_requests: int = 100,
        connector: Optional[aiohttp.TCPConnector] = None,
        connector_owner: Optional[bool] = True,
    ) -> CSelf:
        """
        Construct an authenticated client using the given token.

        See `aiochris.client.base.BaseChrisClient.new` for parameter documentation.
        """
        return await cls.new(
            url=url,
            max_search_requests=max_search_requests,
            connector=connector,
            connector_owner=connector_owner,
            session_modifier=cls.__curry_token(token),
        )

    @classmethod
    async def from_chrs(
        cls,
        url: Optional[str | ChrisURL] = None,
        username: Optional[str | Username] = None,
        max_search_requests: int = 100,
        connector: Optional[aiohttp.TCPConnector] = None,
        connector_owner: Optional[bool] = True,
        config_file: Path = Path("~/.config/chrs/login.toml"),
    ) -> CSelf:
        """
        Log in using [`chrs`](https://crates.io/crates/chrs).
        *ChRIS* logins can be saved with the `chrs login` command.

        In order to call this function, `aiochris` must be installed with the extras `from-chrs`.
        Using pip:

        ```shell
        pip install aiochris[chrs]
        ```

        Or using Poetry:

        ```shell
        poetry add -E chrs aiochris
        ```

        `from_chrs` makes it easy to use `aiochris` in Jupyter Notebook or IPython,
        especially since it saves you from having to write your password in a notebook
        that you want to share with others. Both Jupyter and IPython support top-level `await`.

        ```python
        from aiochris import ChrisClient, acollect

        chris = await ChrisClient.from_chrs()
        await acollect(chris.search_plugins())
        ```

        When `from_chrs` is called with no parameters, it uses the "preferred account"
        i.e. the most recently added account, the same _ChRIS_ account and server as
        `chrs` would when called without options. The "preferred account" can be changed
        by running `chrs switch`.
        """
        logins = ChrsLogins.load(config_file)
        if (t := logins.get_token_for(url, username)) is None:
            raise IncorrectLoginError("No chrs login found.")
        url, token = t
        return await cls.from_token(
            url=url,
            token=token,
            max_search_requests=max_search_requests,
            connector=connector,
            connector_owner=connector_owner,
        )

    @staticmethod
    def __curry_token(token: str) -> Callable[[aiohttp.ClientSession], None]:
        def add_token_to(session: aiohttp.ClientSession) -> None:
            session.headers.update({"Authorization": "Token " + token})

        return add_token_to

    # ============================================================
    # CUBE API methods
    # ============================================================

    @http.search(".")
    def search_feeds(self, **query) -> Search[Feed]:
        """
        Search for feeds.
        """
        ...

    @http.search("plugins")
    def search_plugins(self, **query) -> Search[Plugin]:
        """
        Search for plugins.
        """
        ...

    @http.search("plugin_instances")
    def plugin_instances(self, **query) -> Search[PluginInstance]:
        """
        Search for plugin instances.
        """
        ...

    async def upload_file(
        self, local_file: str | os.PathLike, upload_path: str
    ) -> File:
        """
        Upload a local file to *ChRIS*.

        .. warning:: Uses non-async code.
                     The file is read using non-async code.
                     Performance will suffer with large files and hard drives.
                     See [aiolibs/aiohttp#7174](https://github.com/aio-libs/aiohttp/issues/7174)

        Examples
        --------

        Upload a single file:

        ```python
        aiochris = await ChrisClient.from_login(
            username='chris',
            password='chris1234',
            url='https://cube.chrisproject.org/api/v1/'
        )
        file = await aiochris.upload_file("./my_data.dat", 'dir/my_data.dat')
        assert file.fname == 'aiochris/uploads/dir/my_data.dat'
        ```

        Upload (in parallel) all `*.txt` files in a directory
        `'incoming'` to `aiochris/uploads/big_folder`:

        ```python
        upload_jobs = (
            aiochris.upload_file(p, f'big_folder/{p}')
            for p in Path('incoming')
        )
        await asyncio.gather(upload_jobs)
        ```

        Parameters
        ----------
        local_file
            Path of an existing local file to upload.
        upload_path
            A subpath of `{username}/uploads/` where to upload the file to in *CUBE*
        """
        upload_path = await self._add_upload_prefix(upload_path)
        local_file = Path(local_file)
        with local_file.open("rb") as f:
            data = aiohttp.FormData()
            data.add_field("upload_path", upload_path)
            data.add_field("fname", f, filename=local_file.name)
            sent = self.s.post(self.collection_links.uploadedfiles, data=data)
            return await deserialize_res(
                sent, self, {"fname": local_file, "upload_path": upload_path}, File
            )

        # read_stream = _file_sender(local_file, chunk_size)
        # file_length = os.stat(local_file).st_size
        # return await self.upload_stream(read_stream, upload_path, str(local_file), file_length)

    # doesn't work: 411 Length Required
    # async def upload_stream(self, read_stream: AsyncIterable[bytes], upload_path: str, fname: str, length: int
    #                         ) -> File:
    #     """
    #     Stream a file upload to *ChRIS*. For a higher-level wrapper which accepts
    #     a path argument instead, see `upload`.
    #
    #     Parameters
    #     ----------
    #     read_stream
    #         bytes stream
    #     upload_path
    #         uploadedfiles path starting with `'{username}/uploads/`
    #     fname
    #         file name to use in the multipart POST request
    #     length
    #         content length
    #     """
    #     data = aiohttp.FormData()
    #     data.add_field('upload_path', upload_path)
    #     data.add_field('fname', read_stream, filename=fname)
    #     async with self.s.post(self.collection_links.uploadedfiles, data=data) as res:
    #         return serde.json.from_json(File, await res.text())
    #
    #     with aiohttp.MultipartWriter() as mpwriter:
    #         mpwriter.append_form({'upload_path': upload_path})
    #         mpwriter.append(read_stream, headers={
    #             'Content-Disposition': 'form-data; name="fname"; filename="value_goes_here"'
    #         })

    async def _add_upload_prefix(self, upload_path: str) -> str:
        upload_prefix = f"{await self.username()}/uploads/"
        if str(upload_path).startswith(upload_prefix):
            return upload_path
        return f"{upload_prefix}{upload_path}"

    @http.get("user")
    async def user(self) -> User:
        """Gets the user's information."""
        ...

    @async_cached_property
    async def _user(self) -> User:
        return await self.user()

    async def username(self) -> Username:
        """
        Gets the username. In contrast to `self.user`, this method will use a cached API call.
        """
        user = await self._user  # this is weird
        return user.username

    @http.search("compute_resources")
    def search_compute_resources(self, **query) -> Search[ComputeResource]:
        """
        Search for existing compute resources.

        See also
        --------
        `get_all_compute_resources` :
        """
        ...

    async def get_all_compute_resources(self) -> Sequence[ComputeResource]:
        """
        Get all compute resources.

        This method exists for convenience.
        The number of compute resources of a CUBE is typically small so it's ok.

        See also
        --------
        `search_compute_resources` :
        """
        return await acollect(self.search_compute_resources())

    @http.search("pacsfiles")
    def search_pacsfiles(self, **query) -> Search[PACSFile]:
        """
        Search for PACS files.
        """
        ...


# async def _file_sender(file_name: str | os.PathLike, chunk_size: int):
#     """
#     Stream the reading of a file using an asynchronous generator.
#
#     https://docs.aiohttp.org/en/stable/client_quickstart.html#streaming-uploads
#     """
#     async with aiofiles.open(file_name, 'rb') as f:
#         chunk = await f.read(chunk_size)
#         while chunk:
#             yield chunk
#             chunk = await f.read(chunk_size)



"""
Read-only models for CUBE resources.
"""
import sys
from dataclasses import dataclass
from typing import Optional, Literal, TextIO, Any

import serde
from serde import deserialize

from aiochris.link import http
from aiochris.link.linked import LinkedModel
from aiochris.enums import PluginType
from aiochris.types import *
from aiochris.util.search import Search


@deserialize
@dataclass(frozen=True)
class ComputeResource:
    url: ApiUrl
    id: ComputeResourceId
    creation_date: str
    modification_date: str
    name: ComputeResourceName
    compute_url: PfconUrl
    compute_auth_url: str
    description: str
    max_job_exec_seconds: int


@deserialize
@dataclass(frozen=True)
class PluginParameter(LinkedModel):
    """
    Information about a parameter (a command-line option/flag) of a plugin.
    """

    url: PluginParameterUrl
    id: PluginParameterId
    name: ParameterName
    type: ParameterType
    optional: bool
    default: Optional[ParameterType]
    flag: str
    short_flag: str
    action: Literal["store", "store_true", "store_false"]
    help: str
    ui_exposed: bool
    plugin: PluginUrl


@deserialize
@dataclass(frozen=True)
class PublicPlugin(LinkedModel):
    """
    A ChRIS plugin.
    """

    url: PluginUrl
    id: PluginId
    name: PluginName
    version: PluginVersion
    dock_image: ImageTag
    public_repo: str
    compute_resources: ComputeResourceUrl
    parameters: PluginParametersUrl
    plugin_type: PluginType = serde.field(rename="type")

    @http.search("compute_resources", subpath="")
    def get_compute_resources(self) -> Search[ComputeResource]:
        """Get the compute resources this plugin is registered to."""
        ...

    @http.search("parameters", subpath="")
    def get_parameters(self) -> Search[PluginParameter]:
        """Get the parameters of this plugin."""
        ...

    async def print_help(self, out: TextIO = sys.stdout) -> None:
        """
        Display the help messages for this plugin's parameters.
        """
        async for param in self.get_parameters():
            left = f"{param.name} ({param.flag})"
            out.write(f"{left:>20}: {param.help}")
            if param.default is not None:
                out.write(f" (default: {param.default})")
            out.write("\n")



import enum


class Status(enum.Enum):
    """
    Possible statuses of a plugin instance.
    """

    created = "created"
    waiting = "waiting"
    scheduled = "scheduled"
    started = "started"
    registeringFiles = "registeringFiles"
    finishedSuccessfully = "finishedSuccessfully"
    finishedWithError = "finishedWithError"
    cancelled = "cancelled"


class ParameterTypeName(enum.Enum):
    """
    Plugin parameter types.
    """

    string = "string"
    integer = "integer"
    float = "float"
    boolean = "boolean"


class PluginType(enum.Enum):
    """
    Plugin type.
    """

    ds = "ds"
    fs = "fs"
    ts = "ts"



import dataclasses
import functools
from dataclasses import dataclass
from typing import Iterator, Optional

from serde import deserialize

from aiochris.types import ApiUrl, UserUrl, AdminUrl


@deserialize
@dataclass(frozen=True)
class AbstractCollectionLinks:
    @classmethod
    def has_field(cls, field_name: str) -> bool:
        return field_name in cls._field_names()

    @classmethod
    def _field_names(cls) -> Iterator[str]:
        return (f.name for f in dataclasses.fields(cls))

    def get(self, collection_name: str) -> str:
        url = self._dict.get(collection_name, None)
        if url is None:
            raise TypeError(
                f'{self.__class__} does not have link for "{collection_name}"'
            )
        return url

    @functools.cached_property
    def _dict(self) -> dict[str, str]:
        return dataclasses.asdict(self)


@deserialize
@dataclass(frozen=True)
class AnonymousCollectionLinks(AbstractCollectionLinks):
    chrisinstance: ApiUrl
    compute_resources: ApiUrl
    plugin_metas: ApiUrl
    plugins: ApiUrl
    plugin_instances: ApiUrl
    pipelines: ApiUrl
    pipeline_instances: ApiUrl
    workflows: ApiUrl
    tags: ApiUrl
    pacsfiles: ApiUrl
    servicefiles: ApiUrl
    filebrowser: ApiUrl


@deserialize
@dataclass(frozen=True)
class CollectionLinks(AnonymousCollectionLinks):
    user: UserUrl
    userfiles: Optional[ApiUrl]
    uploadedfiles: Optional[ApiUrl]

    def __post_init__(self):
        if not ((self.userfiles is None) ^ (self.uploadedfiles is None)):
            raise ValueError("Either userfiles or uploadedfiles link must be present")

    @property
    def useruploadedfiles(self) -> ApiUrl:
        if self.userfiles is None:
            return self.uploadedfiles
        return self.userfiles


@deserialize
@dataclass(frozen=True)
class AdminCollectionLinks(CollectionLinks):
    admin: AdminUrl


@deserialize
@dataclass(frozen=True)
class AdminApiCollectionLinks(AbstractCollectionLinks):
    compute_resources: ApiUrl



import dataclasses
from typing import TypeVar, Generic, Type

import yarl

from aiochris.link.linked import Linked
from aiochris.link.metaprog import generic_of
from aiochris.models.collection_links import AbstractCollectionLinks

L = TypeVar("L", bound=AbstractCollectionLinks)


@dataclasses.dataclass(frozen=True)
class CollectionJsonApiClient(Generic[L], Linked):
    """
    A class base for HTTP clients which accept "application/json" from a Collection+JSON API.

    The special value "." is recognized as a way to refer to the collection of the base URL.
    For *CUBE*, it would get the feeds.
    """

    url: str
    """Base API URL"""
    collection_links: L
    """Base API collection links"""

    def _get_link(self, name: str) -> yarl.URL:
        if name == ".":
            return yarl.URL(self.url)
        link = self.collection_links.get(name)
        return yarl.URL(link)

    @classmethod
    def _has_link(cls, name: str) -> bool:
        return name == "." or cls._collection_type().has_field(name)

    @classmethod
    def _collection_type(cls) -> Type[L]:
        return generic_of(cls, AbstractCollectionLinks)



"""
.. include:: ./home.md

.. include:: ./examples.md
"""

import aiochris.client
import aiochris.models
import aiochris.util
from aiochris.util.search import Search, acollect
from aiochris.client.normal import ChrisClient
from aiochris.client.anon import AnonChrisClient
from aiochris.client.admin import ChrisAdminClient
from aiochris.enums import Status, ParameterTypeName

__all__ = [
    "AnonChrisClient",
    "ChrisClient",
    "ChrisAdminClient",
    "Search",
    "acollect",
    "Status",
    "ParameterTypeName",
    "client",
    "models",
    "util",
    "errors",
    "types",
]



"""
Subclasses of classes from `aiochris.models.data` which are returned
from an `aiochris.client.authed.AuthenticatedClient`.
These classes may have read-write functionality on the *ChRIS* API.
"""
import asyncio
import time
from dataclasses import dataclass
from typing import Optional, Sequence

from serde import deserialize

from aiochris.link import http
from aiochris.link.linked import LinkedModel
from aiochris.models.data import PluginInstanceData, FeedData, UserData, FeedNoteData
from aiochris.enums import PluginType, Status
from aiochris.models.public import PublicPlugin
from aiochris.types import *


@deserialize
@dataclass(frozen=True)
class User(UserData, LinkedModel):
    pass  # TODO change_email, change_password


@deserialize
@dataclass(frozen=True)
class File(LinkedModel):
    """
    A file in CUBE.
    """

    url: str
    fname: FileFname
    fsize: int
    file_resource: FileResourceUrl

    @property
    def parent(self) -> str:
        """
        Get the parent (directory) of a file.

        Examples
        --------

        ```python
        assert file.fname == 'chris/feed_4/pl-dircopy_7/data/hello-world.txt'
        assert file.parent == 'chris/feed_4/pl-dircopy_7/data'
        ```
        """
        split = self.fname.split("/")
        if len(split) <= 1:
            return self.fname
        return "/".join(split[:-1])

    # TODO download methods


@deserialize
@dataclass(frozen=True)
class PACSFile(File):
    """
    A file from a PACS server which was pushed into ChRIS.
    A PACSFile is usually a DICOM file.

    See https://github.com/FNNDSC/ChRIS_ultron_backEnd/blob/a1bf499144df79622eb3f8a459cdb80d8e34cb04/chris_backend/pacsfiles/models.py#L16-L33
    """

    PatientID: str
    PatientName: str
    PatientBirthDate: str
    PatientAge: int
    PatientSex: str
    StudyDate: str  # TODO deserialize
    AccessionNumber: str
    Modality: str
    ProtocolName: str
    StudyInstanceUID: str
    StudyDescription: str
    SeriesInstanceUID: str
    SeriesDescription: str
    pacs_identifier: str


@deserialize
@dataclass(frozen=True)
class PluginInstance(PluginInstanceData):
    @http.get("feed")
    async def get_feed(self) -> "Feed":
        """Get the feed this plugin instance belongs to."""
        ...

    @http.put("url")
    async def get(self) -> "PluginInstance":
        """
        Get this plugin's state (again).
        """
        ...

    @http.put("url")
    async def set(
        self, title: Optional[str] = None, status: Optional[str] = None
    ) -> "PluginInstance":
        """
        Set the title or status of this plugin instance.
        """
        ...

    @http.delete("url")
    async def delete(self) -> None:
        """Delete this plugin instance."""
        ...

    async def wait(
        self,
        status: Status
        | Sequence[Status] = (
            Status.finishedSuccessfully,
            Status.finishedWithError,
            Status.cancelled,
        ),
        timeout: float = 300,
        interval: float = 5,
    ) -> tuple[float, "PluginInstance"]:
        """
        Wait until this plugin instance finishes (or some other desired status).

        Parameters
        ----------
        status
            Statuses to wait for
        timeout
            Number of seconds to wait for before giving up
        interval
            Number of seconds to wait between checking on status

        Returns
        -------
        Returns the number of seconds elapsed and the last state of the plugin instance.
        This function will return for one of two reasons: either the plugin instance finished,
        or this function timed out. Make sure you check the plugin instance's final status!
        """
        if status is Status:
            status = (status,)
        if self.status in status:
            return 0.0, self
        timeout_ns = timeout * 1e9
        start = time.monotonic_ns()
        while (cur := await self.get()).status not in status:
            elapsed = time.monotonic_ns() - start
            if elapsed > timeout_ns:
                return elapsed / 1e9, cur
            await asyncio.sleep(interval)
        return (time.monotonic_ns() - start) / 1e9, cur


@deserialize
@dataclass(frozen=True)
class FeedNote(FeedNoteData):
    @http.get("feed")
    async def get_feed(self) -> "Feed":
        """Get the feed this note belongs to."""
        ...

    @http.put("url")
    async def set(
        self, title: Optional[str] = None, content: Optional[str] = None
    ) -> "FeedNote":
        """Change this note."""
        ...


@deserialize
@dataclass(frozen=True)
class Feed(FeedData):
    """
    A feed of a logged in user.
    """

    @http.put("url")
    async def set(
        self, name: Optional[str] = None, owner: Optional[str | Username] = None
    ) -> "Feed":
        """
        Change the name or the owner of this feed.

        Parameters
        ----------
        name
            new name for this feed
        owner
            new owner for this feed
        """
        ...

    @http.get("note")
    async def get_note(self) -> FeedNote:
        ...


@deserialize
@dataclass(frozen=True)
class Plugin(PublicPlugin):
    """
    A ChRIS plugin. Create a plugin instance of this plugin to run it.
    """

    instances: ApiUrl

    @http.post("instances")
    async def _create_instance_raw(self, **kwargs) -> PluginInstance:
        ...

    async def create_instance(
        self, previous: Optional[PluginInstance] = None, **kwargs
    ) -> PluginInstance:
        """
        Create a plugin instance, i.e. "run" this plugin.

        Parameters common to all plugins are described below.
        Not all valid parameters are listed, since each plugin's parameters are different.
        Some plugins have required parameters too.
        To list all possible parameters, make a GET request to the specific plugin's instances link.

        Parameters
        ----------
        previous: chris.models.data.PluginInstanceData
            Previous plugin instance
        previous_id: int
            Previous plugin instance ID number (conflicts with `previous`)
        compute_resource_name: Optional[str]
            Name of compute resource to use
        memory_limit: Optional[str]
            Memory limit. Format is *x*Mi or *x*Gi where x is an integer.
        cpu_limit: Optional[str]
            CPU limit. Format is *x*m for *x* is an integer in millicores.
        gpu_limit: Optional[int]
            GPU limit.
        """
        if previous is not None:
            if "previous_id" in kwargs:
                raise ValueError("Cannot give both previous and previous_id.")
            if not isinstance(previous, PluginInstance):
                raise TypeError(f"{previous} is not a PluginInstance")
            kwargs["previous_id"] = previous.id
        if self.plugin_type is PluginType.fs:
            if "previous_id" in kwargs:
                raise ValueError(
                    "Cannot create plugin instance of a fs-type plugin with a previous plugin instance."
                )
        elif "previous_id" not in kwargs:
            raise ValueError(
                f'Plugin type is "{self.plugin_type.value}" so previous is a required parameter.'
            )
        return await self._create_instance_raw(**kwargs)



"""
Interoperability with [`chrs`](https://crates.io/crates/chrs) version 0.2.4.

The functions of this module uses dynamic importing of extras from the `chrs` group.
"""
import dataclasses
from pathlib import Path

import serde
from typing import Optional, Literal
from aiochris.types import Username, ChrisURL
import functools

_SERVICE = "org.chrisproject.chrs"
"""
Keyring service name used by `chrs`.

https://github.com/FNNDSC/chrs/blob/v0.2.4/chrs/src/login/saved.rs#L14
"""


@functools.cache
def _get_keyring():
    import keyring

    return keyring.get_keyring()


@serde.deserialize
@dataclasses.dataclass(frozen=True)
class StoredToken:
    """
    https://github.com/FNNDSC/chrs/blob/v0.2.4/chrs/src/login/tokenstore.rs#L18-L24
    """

    # note: we can't do @serde.deserialize(tagging=serde.AdjacentTagging('store', 'value'))
    # See https://github.com/yukinarit/pyserde/issues/411
    store: Literal["Text", "Keyring"]
    value: Optional[str] = None

    def __post_init__(self):
        if self.store not in ("Text", "Keyring"):
            raise ValueError()
        if self.store == "Text" and self.store is None:
            raise ValueError()


@dataclasses.dataclass(frozen=True)
class ChrsLogin:
    """
    A login saved by `chrs`.

    https://github.com/FNNDSC/chrs/blob/v0.2.4/chrs/src/login/tokenstore.rs#L18-L34
    """

    address: ChrisURL
    username: Username
    store: StoredToken

    def token(self) -> str:
        if self.store.value is not None:
            return self.store.value
        return self._get_token_from_keyring()

    def is_for(self, address: Optional[ChrisURL], username: Optional[Username]) -> bool:
        """
        Returns `True` if this login is for the specified _ChRIS_ user account.
        """
        if address is None and username is None:
            return True
        if address is None and username == self.username:
            return True
        if username is None and address == self.address:
            return True
        if address == self.address and username == self.username:
            return True
        return False

    def to_keyring_username(self) -> str:
        """
        Produce the username for this login in the keyring.

        https://github.com/FNNDSC/chrs/blob/v0.2.4/chrs/src/login/tokenstore.rs#L3
        """
        return f"{self.username}@{self.address}"

    def _get_token_from_keyring(self) -> str:
        """
        https://github.com/FNNDSC/chrs/blob/v0.2.4/chrs/src/login/tokenstore.rs#L110-L112
        """
        if self.store.store != "Keyring":
            raise ChrsKeyringError(
                "chrs login config is not valid: "
                f"for address={self.address} and username={self.username}"
                f"expected store=Keyring, got store={self.store.store}"
            )
        keyring_username = self.to_keyring_username()
        token = _get_keyring().get_password(_SERVICE, keyring_username)
        if token is None:
            raise ChrsKeyringError(f"No keyring password for {keyring_username}")
        return token


@dataclasses.dataclass(frozen=True)
class ChrsLogins:
    """
    Logins saved by `chrs`.

    https://github.com/FNNDSC/chrs/blob/v0.2.4/chrs/src/login/saved.rs#L18-L22
    """

    cubes: list[ChrsLogin]
    """Saved logins in reverse order of preference."""

    @classmethod
    def load(cls, path: Path) -> "ChrsLogins":
        import serde.toml

        return serde.toml.from_toml(cls, path.expanduser().read_text())

    def get_token_for(
        self,
        address: Optional[str | ChrisURL] = None,
        username: Optional[str | Username] = None,
    ) -> Optional[tuple[ChrisURL, str]]:
        """
        Get token for a login.
        """
        for login in reversed(self.cubes):
            if login.is_for(address, username):
                if (token := login.token()) is not None:
                    return login.address, token
        return None


class ChrsKeyringError(Exception):
    """Error interacting with Keyring."""

    pass



from typing import Optional

import aiohttp

from aiochris.client.base import BaseChrisClient
from aiochris.link import http
from aiochris.models.collection_links import AnonymousCollectionLinks
from aiochris.models.public import PublicPlugin
from aiochris.util.search import Search


class AnonChrisClient(BaseChrisClient[AnonymousCollectionLinks, "AnonChrisClient"]):
    """
    An anonymous ChRIS client. It has access to read-only GET resources,
    such as being able to search for plugins.
    """

    @classmethod
    async def from_url(
        cls,
        url: str,
        max_search_requests: int = 100,
        connector: Optional[aiohttp.BaseConnector] = None,
        connector_owner: bool = True,
    ) -> "AnonChrisClient":
        """
        Create an anonymous client.

        See `aiochris.client.base.BaseChrisClient.new` for parameter documentation.
        """
        return await cls.new(
            url=url,
            max_search_requests=max_search_requests,
            connector=connector,
            connector_owner=connector_owner,
        )

    @http.search("plugins")
    def search_plugins(self, **query) -> Search[PublicPlugin]:
        """
        Search for plugins.

        Since this client is not logged in, you cannot create plugin instances using this client.
        """
        ...



"""
Dataclasses describing objects returned from CUBE.

These classes are extended in the modules `aiochris.models.logged_in`
and `aiochris.models.public` with methods to get objects from links.
"""

from dataclasses import dataclass
import datetime
from typing import Optional

from serde import deserialize

from aiochris.link.linked import LinkedModel
from aiochris.enums import PluginType, Status
from aiochris.types import *


@deserialize
@dataclass(frozen=True)
class UserData:
    """A *CUBE* user."""

    url: UserUrl
    id: UserId
    username: Username
    email: str


# TODO It'd be better to use inheritance instead of optionals
@deserialize
@dataclass(frozen=True)
class PluginInstanceData(LinkedModel):
    """
    A *plugin instance* in _ChRIS_ is a computing job, i.e. an attempt to run
    a computation (a non-interactive command-line app) to produce data.
    """

    url: PluginInstanceUrl
    id: PluginInstanceId
    title: str
    compute_resource_name: ComputeResourceName
    plugin_id: PluginId
    plugin_name: PluginName
    plugin_version: PluginVersion
    plugin_type: PluginType

    pipeline_inst: Optional[int]
    feed_id: FeedId
    start_date: datetime.datetime
    end_date: datetime.datetime
    output_path: CubeFilePath

    status: Status

    summary: str
    raw: str
    owner_username: Username
    cpu_limit: int
    memory_limit: int
    number_of_workers: int
    gpu_limit: int
    error_code: CUBEErrorCode

    previous: Optional[PluginInstanceUrl]
    feed: FeedUrl
    plugin: PluginUrl
    descendants: DescendantsUrl
    files: FilesUrl
    parameters: PluginInstanceParamtersUrl
    compute_resource: ComputeResourceUrl
    splits: SplitsUrl

    previous_id: Optional[int] = None
    """
    FS plugins will not produce a `previous_id` value
    (even though they will return `"previous": null`)
    """

    size: Optional[int] = None
    """
    IDK what it is the size of.

    This field shows up when the plugin instance is maybe done,
    but not when the plugin instance is created.
    """
    template: Optional[dict] = None
    """
    Present only when getting a plugin instance.
    """


@deserialize
@dataclass(frozen=True)
class FeedData(LinkedModel):
    url: FeedUrl
    id: FeedId
    creation_date: datetime.datetime
    modification_date: datetime.datetime
    name: str
    creator_username: Username
    created_jobs: int
    waiting_jobs: int
    scheduled_jobs: int
    started_jobs: int
    registering_jobs: int
    finished_jobs: int
    errored_jobs: int
    cancelled_jobs: int
    owner: list[UserUrl]
    note: NoteUrl
    tags: TagsUrl
    taggings: TaggingsUrl
    comments: CommentsUrl
    files: FilesUrl
    plugin_instances: PluginInstancesUrl


@deserialize
@dataclass(frozen=True)
class FeedNoteData(LinkedModel):
    url: FeedUrl
    id: NoteId
    title: str
    content: str
    feed: FeedUrl



import io
import json
from typing import Iterable

import aiohttp
from async_property import async_cached_property
from serde import from_dict

from aiochris.client.authed import AuthenticatedClient
from aiochris.link import http
from aiochris.link.collection_client import CollectionJsonApiClient
from aiochris.link.linked import deserialize_linked
from aiochris.models.collection_links import (
    AdminCollectionLinks,
    AdminApiCollectionLinks,
)
from aiochris.models.logged_in import Plugin
from aiochris.models.public import ComputeResource
from aiochris.types import PluginUrl, ComputeResourceName, PfconUrl
from aiochris.errors import raise_for_status


class _AdminApiClient(CollectionJsonApiClient[AdminApiCollectionLinks]):
    """
    A client to `/chris-admin/api/v1/`
    """

    @http.post("compute_resources")
    async def create_compute_resource(self, **kwargs) -> ComputeResource:
        ...


class ChrisAdminClient(AuthenticatedClient[AdminCollectionLinks, "ChrisAdminClient"]):
    """
    A client who has access to `/chris-admin/`. Admins can register new plugins and
    add new compute resources.
    """

    @http.post("admin")
    async def _register_plugin_from_store_raw(
        self, plugin_store_url: str, compute_names: str
    ) -> Plugin:
        ...

    async def register_plugin_from_store(
        self, plugin_store_url: PluginUrl, compute_names: Iterable[ComputeResourceName]
    ) -> Plugin:
        """
        Register a plugin from a ChRIS Store.
        """
        return await self._register_plugin_from_store_raw(
            plugin_store_url=plugin_store_url, compute_names=",".join(compute_names)
        )

    async def add_plugin(
        self,
        plugin_description: str | dict,
        compute_resources: str
        | ComputeResource
        | Iterable[ComputeResource | ComputeResourceName],
    ) -> Plugin:
        """
        Add a plugin to *CUBE*.

        Examples
        --------

        ```python
        cmd = ['docker', 'run', '--rm', 'fnndsc/pl-mri-preview', 'chris_plugin_info']
        output = subprocess.check_output(cmd, text=True)
        desc = json.loads(output)
        desc['name'] = 'pl-mri-preview'
        desc['public_repo'] = 'https://github.com/FNNDSC/pl-mri-preview'
        desc['dock_image'] = 'fnndsc/pl-mri-preview'

        await chris_admin.add_plugin(plugin_description=desc, compute_resources='host')
        ```

        The example above is just for show. It's not a good example for several reasons:

        - Calls blocking function `subprocess.check_output` in asynchronous context
        - It is preferred to use a versioned string for `dock_image`
        - `host` compute environment is not guaranteed to exist. Instead, you could
          call `aiochris.client.authed.AuthenticatedClient.search_compute_resources`
          or `aiochris.client.authed.AuthenticatedClient.get_all_compute_resources`:

        ```python
        all_computes = await chris_admin.get_all_compute_resources()
        await chris_admin.add_plugin(plugin_description=desc, compute_resources=all_computes)
        ```

        Parameters
        ----------
        plugin_description: str | dict
            JSON description of a plugin.
            [spec](https://github.com/FNNDSC/CHRIS_docs/blob/5078aaf934bdbe313e85367f88aff7c14730a1d4/specs/ChRIS_Plugins.adoc#descriptor_file)
        compute_resources
            Compute resources to register the plugin to. Value can be either a comma-separated `str` of names,
            a `aiochris.models.public.ComputeResource`, a sequence of `aiochris.models.public.ComputeResource`,
            or a sequence of compute resource names as `str`.
        """
        compute_names = _serialize_crs(compute_resources)
        if not isinstance(plugin_description, str):
            plugin_description = json.dumps(plugin_description)
        data = aiohttp.FormData()
        data.add_field(
            "fname",
            io.StringIO(plugin_description),
            filename="aiochris_add_plugin.json",
        )
        data.add_field("compute_names", compute_names)
        async with self.s.post(self.collection_links.admin, data=data) as res:
            await raise_for_status(res)
            return deserialize_linked(self, Plugin, await res.json())

    async def create_compute_resource(
        self,
        name: str | ComputeResourceName,
        compute_url: str | PfconUrl,
        compute_user: str,
        compute_password: str,
        compute_innetwork: bool = None,
        description: str = None,
        compute_auth_url: str = None,
        compute_auth_token: str = None,
        max_job_exec_seconds: str = None,
    ) -> ComputeResource:
        """
        Define a new compute resource.
        """
        return await (await self._admin).create_compute_resource(
            name=name,
            compute_url=compute_url,
            compute_user=compute_user,
            compute_password=compute_password,
            compute_innetwork=compute_innetwork,
            description=description,
            compute_auth_url=compute_auth_url,
            compute_auth_token=compute_auth_token,
            max_job_exec_seconds=max_job_exec_seconds,
        )

    @async_cached_property
    async def _admin(self) -> _AdminApiClient:
        """
        Get a (sub-)client for `/chris-admin/api/v1/`
        """
        res = await self.s.get(self.collection_links.admin)
        body = await res.json()
        links = from_dict(AdminApiCollectionLinks, body["collection_links"])
        return _AdminApiClient(
            url=self.collection_links.admin,
            s=self.s,
            collection_links=links,
            max_search_requests=self.max_search_requests,
        )


def _serialize_crs(
    c: str | ComputeResource | Iterable[ComputeResource | ComputeResourceName],
) -> str:
    if isinstance(c, str):
        return c
    if isinstance(c, ComputeResource):
        return c.name
    if not isinstance(c, Iterable):
        raise TypeError("compute_resources must be str or Iterable")
    return ",".join(map(_serialize_cr, c))


def _serialize_cr(c: str | ComputeResource):
    if isinstance(c, ComputeResource):
        return c.name
    return str(c)



import aiohttp


async def raise_for_status(res: aiohttp.ClientResponse) -> None:
    """
    Raises custom exceptions.
    """
    if res.status < 400:
        res.raise_for_status()
        return
    exception = BadRequestError if res.status < 500 else InternalServerError
    try:
        raise exception(res.status, res.url, await res.json())
    except aiohttp.ClientError:
        raise exception(res.status, res.url)


class BaseClientError(Exception):
    pass


class ResponseError(BaseClientError):
    pass


class BadRequestError(ResponseError):
    pass


class InternalServerError(ResponseError):
    pass


class IncorrectLoginError(BaseClientError):
    pass


class NonsenseResponseError(ResponseError):
    """CUBE returned data which does not make sense."""

    pass



"""
Metaprogramming helper functions.
"""
import abc
import functools
import typing
from typing import Type, Callable, TypeVar, ForwardRef, Optional

import typing_inspect

_T = TypeVar("_T")


def get_return_hint(fn: Callable[[...], _T]) -> Type[_T]:
    hints = typing.get_type_hints(fn)
    if "return" not in hints:
        raise ValueError(f"Function {fn} must define a return type hint.")
    return hints["return"]


@functools.cache
def generic_of(c: Type, t: Type[_T], is_subclass=False) -> Optional[Type[_T]]:
    """
    Get the actual class represented by a bound TypeVar of a generic.
    """
    for generic_type in typing_inspect.get_args(c):
        if isinstance(generic_type, (str, ForwardRef, TypeVar)):
            continue
        if issubclass(generic_type, t):
            return generic_type

    if hasattr(c, "__orig_bases__"):
        for subclass in c.__orig_bases__:
            if subclass is abc.ABC:
                return None
            subclass_generic = generic_of(subclass, t, is_subclass=True)
            if subclass_generic is not None:
                return subclass_generic
    if not is_subclass:
        raise TypeError(
            f"No generic of {t} found in {c} nor its subclasses."
            "\nWARNING: generic_of fails if importlib.reload was called "
            "to reload the generic type's superclass. This happens "
            "when running pdoc as a webserver."
        )
    return None



"""
NewTypes for ChRIS models.
"""

from typing import NewType, Union

Username = NewType("Username", str)
"""ChRIS user account username"""
Password = NewType("Password", str)
"""ChRIS user account password"""
ChrisURL = NewType("ChrisURL", str)
"""ChRIS backend API base URL"""

ApiUrl = NewType("ApiUrl", str)
"""Any CUBE URL which I am too lazy to be more specific about."""
ResourceId = NewType("ResourceId", int)
"""ID number which I am too lazy to be more specific about."""
PluginName = NewType("PluginName", str)
"""Name of a ChRIS plugin"""
ImageTag = NewType("ImageTag", str)
"""OCI image tag (informally, a Docker Image name)"""
PluginVersion = NewType("PluginVersion", str)
"""Version string of a ChRIS plugin"""

PluginUrl = NewType("PluginUrl", str)
"""
URL of a ChRIS plugin.

## Examples

- https://chrisstore.co/api/v1/plugins/5/
- https://cube.chrisproject.org/api/v1/plugins/6/
"""
PluginSearchUrl = NewType("PluginSearchUrl", str)

PluginId = NewType("PluginId", int)

UserUrl = NewType("UserUrl", str)
UserId = NewType("UserId", int)

AdminUrl = NewType("AdminUrl", str)
"""A admin resource URL ending with `/chris-admin/api/v1/`"""
ComputeResourceName = NewType("ComputeResourceName", str)
ComputeResourceId = NewType("ComputeResourceId", str)

PfconUrl = NewType("PfconUrl", str)

FeedId = NewType("FeedId", str)


CubeFilePath = NewType("CubeFilePath", str)


CUBEErrorCode = NewType("CUBEErrorCode", str)

ContainerImageTag = NewType("ContainerImageTag", str)

PipingId = NewType("PipingId", int)
PipelineId = NewType("PipelineId", int)


ParameterName = NewType("ParameterName", str)
ParameterType = Union[str, int, float, bool]

PipelineParameterId = NewType("ParameterLocalId", int)
PluginParameterId = NewType("ParameterGlobalId", int)
PluginInstanceId = NewType("PluginInstanceId", int)

FileFname = NewType("FileFname", str)
FileResourceName = NewType("FileResourceName", str)
FileId = NewType("FileId", int)

FilesUrl = NewType("FilesUrl", str)
FileResourceUrl = NewType("FileResourceUrl", str)
PipelineUrl = NewType("PipelineUrl", str)
PipingsUrl = NewType("PipingsUrl", str)
PipelinePluginsUrl = NewType("PipelinePluginsUrl", str)
PipelineDefaultParametersUrl = NewType("PipelineDefaultParametersUrl", str)
PipingUrl = NewType("PipingUrl", str)

PipelineParameterUrl = NewType("PipingParameterUrl", str)
PluginInstanceUrl = NewType("PluginInstanceUrl", str)
PluginInstancesUrl = NewType("PluginInstancesUrl", str)
DescendantsUrl = NewType("DescendantsUrl", str)
PipelineInstancesUrl = NewType("PipelineInstancesUrl", str)
PluginInstanceParamtersUrl = NewType("PluginInstanceParametersUrl", str)
ComputeResourceUrl = NewType("ComputeResourceUrl", str)
SplitsUrl = NewType("SplitsUrl", str)
FeedUrl = NewType("FeedUrl", str)
NoteId = NewType("NoteId", int)
"""A feed note's ID number."""
NoteUrl = NewType("NoteUrl", str)
"""
A feed's note URL.

## Examples

- https://cube.chrisproject.org/api/v1/note4/
"""
PluginParametersUrl = NewType("PluginParametersUrl", str)
TagsUrl = NewType("TagsUrl", str)
TaggingsUrl = NewType("TaggingsUrl", str)
CommentsUrl = NewType("CommentsUrl", str)

PluginParameterUrl = NewType("PluginParameterUrl", str)



import abc
import dataclasses
from typing import (
    Final,
    Any,
    Optional,
    Callable,
    Type,
    TypeGuard,
    TypeVar,
    AsyncContextManager,
)

import aiohttp
import serde
import yarl

from aiochris.errors import raise_for_status, ResponseError

T = TypeVar("T")


class LinkedMeta(abc.ABCMeta):
    """
    A metaclass for HTTP clients which make requests to URIs
    that come from the API.
    """

    __DECORATED_METHOD_MARK: Final = "_link_name"
    """
    An attribute name which is set on functions decorated by the
    method decorators in this module.
    """

    def __new__(mcs, name, bases, dct):
        c = super().__new__(mcs, name, bases, dct)
        if abc.ABC in bases:
            return c

        if not mcs._is_linked(c):
            raise TypeError(f"{c} must be a subclass of {Linked}")

        mcs._validate_collection_names(c, dct)
        return c

    @classmethod
    def _is_linked(cls, c) -> TypeGuard[Type["Linked"]]:
        return issubclass(c, Linked)

    @classmethod
    def _validate_collection_names(mcs, cls: Type["Linked"], dct: dict[str, Any]):
        """
        Check that every method marked by `mark_to_check` was given a collection name
        which exists as a field in the class's associated
        `aiochris.models.collection_links.AbstractCollectionLinks` type.
        """
        marked_methods = filter(None, map(mcs._get_marked, dct.values()))
        for collection_name, method in marked_methods:
            if not cls._has_link(collection_name):
                raise TypeError(
                    f'Method {method} needs link "{collection_name}" '
                    f"but {cls} does not have it"
                )

    @classmethod
    def _get_marked(mcs, method: Any) -> Optional[tuple[str, Callable]]:
        if not callable(method):
            return None
        if hasattr(method, mcs.__DECORATED_METHOD_MARK):
            link_name = getattr(method, mcs.__DECORATED_METHOD_MARK)
            return link_name, method
        return None

    @classmethod
    def mark_to_check(mcs, method: Callable, link_name: str) -> None:
        """
        Indicate that the given method has a data dependency on a link.
        """
        if not callable(method):
            raise TypeError("method must be callable")
        setattr(method, mcs.__DECORATED_METHOD_MARK, link_name)


@serde.deserialize
@dataclasses.dataclass(frozen=True)
class Linked(abc.ABC, metaclass=LinkedMeta):
    """
    A `Linked` is an object which can make HTTP requests to links from an API.
    """

    s: aiohttp.ClientSession = serde.field(deserializer=lambda s: s)
    max_search_requests: int
    """
    Maximum number of requests to make for pagination.
    """

    @classmethod
    @abc.abstractmethod
    def _has_link(cls, name: str) -> bool:
        ...

    @abc.abstractmethod
    def _get_link(self, name: str) -> yarl.URL:
        ...


@serde.deserialize
@dataclasses.dataclass(frozen=True)
class LinkedModel(Linked, abc.ABC):
    """
    A class where its fields may be API links.
    """

    @classmethod
    def _has_link(cls, name: str) -> bool:
        return any(name == field for field in cls._field_names())

    @classmethod
    def _field_names(cls) -> frozenset[str]:
        # LinkedMeta.__new__ is called before the dataclasses.dataclass
        # decorator, so dataclasses.fields only returns the superclass fields.
        # Hence, it is necessary to also check __annotations__
        return frozenset(
            (*(f.name for f in dataclasses.fields(cls)), *cls.__annotations__.keys())
        )

    def _get_link(self, name: str) -> yarl.URL:
        return yarl.URL(getattr(self, name))


def deserialize_linked(client: Linked, t: Type[T], o: dict) -> T:
    """
    Wraps `serde.from_dict`.
    If `t` is a subclass of `aiochris.models.connected.Connected`, its session is set.

    Side effect: o is mutated
    """
    if issubclass(t, LinkedModel):
        o["s"] = client.s
        o["max_search_requests"] = client.max_search_requests
    return serde.from_dict(t, o, reuse_instances=True)


async def deserialize_res(
    sent_request: AsyncContextManager[aiohttp.ClientResponse],
    client: Linked,
    sent_data: dict,
    return_type: Type[T],
) -> T:
    async with sent_request as res:
        try:
            await raise_for_status(res)
        except ResponseError as e:
            raise e.__class__(*e.args, f"data={sent_data}")
        if return_type is type(None):  # noqa
            return None
        sent_data = await res.json(content_type="application/json")
    return deserialize_linked(client, return_type, sent_data)



import copy
import logging
from dataclasses import dataclass
from typing import (
    Optional,
    TypeVar,
    AsyncGenerator,
    Type,
    AsyncIterable,
    Any,
    Generic,
    AsyncIterator,
)

import yarl
from serde import deserialize
from serde.json import from_json

from aiochris.link.linked import deserialize_linked, Linked
from aiochris.errors import (
    BaseClientError,
    raise_for_status,
    NonsenseResponseError,
)

logger = logging.getLogger(__name__)

T = TypeVar("T")


@deserialize
class _Paginated:
    """
    Response from a paginated endpoint.
    """

    count: int
    next: Optional[str]
    previous: Optional[str]
    results: list[Any]


@dataclass
class Search(Generic[T], AsyncIterable[T]):
    """
    Abstraction over paginated collection responses from *CUBE*.
    `Search` objects are returned by methods for search endpoints of the *CUBE* API.
    It is an [asynchronous iterable](https://docs.python.org/3/glossary.html#term-asynchronous-iterable)
    which produces items from responses that return multiple results.
    HTTP requests are fired as-neede, they happen in the background during iteration.
    No request is made before the first time a `Search` object is called.

    .. note:: Pagination is handled internally and automatically.
             The query parameters `limit` and `offset` can be explicitly given, but they shouldn't.

    Examples
    --------

    Use an `async for` loop to print the name of every feed:

    ```python
    all_feeds = chris.search_feeds()  # returns a Search[Feed]
    async for feed in all_feeds:
        print(feed.name)
    ```
    """

    base_url: str
    params: dict[str, Any]
    client: Linked
    Item: Type[T]
    max_requests: int = 100
    subpath: str = "search/"

    def __aiter__(self) -> AsyncIterator[T]:
        return self._paginate(self.url)

    async def first(self) -> Optional[T]:
        """
        Get the first item.

        See also
        --------
        `get_only` : similar use, but more strict
        """
        return await anext(self._first_aiter(), None)

    async def get_only(self, allow_multiple=False) -> T:
        """
        Get the *only* item from a search with one result.

        Examples
        --------

        This method is very commonly used for getting "one thing" from CUBE.

        ```python
        await chris.search_plugins(name_exact="pl-dircopy", version="2.1.1").get_only()
        ```

        In the example above, a search for plugins given (`name_exact`, `version`)
        is guaranteed to return either 0 or 1 result.

        Raises
        ------
        aiochris.util.search.NoneSearchError
            If this search is empty.
        aiochris.util.search.ManySearchError
            If this search has more than one item and `allow_multiple` is `False`

        See also
        --------
        `first` : does the same thing but without checks.

        Parameters
        ----------
        allow_multiple: bool
            if `True`, do not raise `ManySearchError` if `count > 1`
        """
        one = await self._get_one()
        if one.count == 0:
            raise NoneSearchError(self.url)
        if not allow_multiple and one.count > 1:
            raise ManySearchError(self.url)
        if len(one.results) < 1:
            raise NonsenseResponseError(
                f"Response has count={one.count} but the results are empty.", one
            )
        return deserialize_linked(self.client, self.Item, one.results[0])

    async def count(self) -> int:
        """
        Get the number of items in this collection search.

        Examples
        --------

        `count` is useful for rendering a progress bar. TODO example with files
        """
        one = await self._get_one()
        return one.count

    async def _get_one(self) -> _Paginated:
        async with self.client.s.get(self._first_url) as res:
            await raise_for_status(res)
            return from_json(_Paginated, await res.text())

    def _paginate(self, url: yarl.URL) -> AsyncIterator[T]:
        return _get_paginated(
            client=self.client,
            url=url,
            item_type=self.Item,
            max_requests=self.max_requests,
        )

    @property
    def url(self) -> yarl.URL:
        return self._search_url_with(self.params)

    def _first_aiter(self) -> AsyncIterator[T]:
        return self._paginate(self._first_url)

    @property
    def _first_url(self) -> yarl.URL:
        params = copy.copy(self.params)
        params["limit"] = 1
        params["offset"] = 0
        return self._search_url_with(params)

    @property
    def _search_url(self) -> yarl.URL:
        return yarl.URL(self.base_url) / self.subpath

    def _search_url_with(self, query: dict[str, Any]):
        return yarl.URL(self._search_url).with_query(query)


async def _get_paginated(
    client: Linked,
    url: yarl.URL | str,
    item_type: Type[T],
    max_requests: int,
) -> AsyncGenerator[T, None]:
    """
    Make HTTP GET requests to a paginated endpoint. Further requests to the
    "next" URL are made in the background as needed.
    """
    logger.debug("GET, max_requests=%d, --> %s", max_requests, url)
    if max_requests != -1 and max_requests == 0:
        raise TooMuchPaginationError(
            f"too many requests made to {url}. "
            f"If this is expected, then pass the argument max_search_requests=-1 to "
            f"the client constructor classmethod."
        )
    async with client.s.get(url) as res:  # N.B. not checking for 4XX, 5XX statuses
        data: _Paginated = from_json(_Paginated, await res.text())
        for element in data.results:
            yield deserialize_linked(client, item_type, element)
    if data.next is not None:
        next_results = _get_paginated(client, data.next, item_type, max_requests - 1)
        async for next_element in next_results:
            yield next_element


async def acollect(async_iterable: AsyncIterable[T]) -> list[T]:
    """
    Simple helper to convert a `Search` to a [`list`](https://docs.python.org/3/library/stdtypes.html#list).

    Using this function is not recommended unless you can assume the collection is small.
    """
    # nb: using tuple here causes
    #     TypeError: 'async_generator' object is not iterable
    # return tuple(e async for e in async_iterable)
    return [e async for e in async_iterable]


class TooMuchPaginationError(BaseClientError):
    """Specified maximum number of requests exceeded while retrieving results from a paginated resource."""

    pass


class GetOnlyError(BaseClientError):
    """Search does not have exactly one result."""

    pass


class NoneSearchError(GetOnlyError):
    """A search expected to have at least one element, has none."""

    pass


class ManySearchError(GetOnlyError):
    """A search expected to have only one result, has several."""

    pass



"""
Decorators for making HTTP requests to API links.

The specified URLs for API endpoints come from the API itself,
so URLs are resolved at the time a method is called.

The decorated method's `**kwargs` are serialized and sent along with the HTTP request,
and the response is deserialized according to the method's return type hint.
"""

import functools
import logging
import typing
from typing import (
    Callable,
    TypeVar,
    Type,
    Any,
    Optional,
    AsyncContextManager,
    Coroutine,
)

import aiohttp
import yarl

from aiochris.link.linked import LinkedMeta, Linked, deserialize_res
from aiochris.link.metaprog import get_return_hint
from aiochris.util.search import Search

logger = logging.getLogger(__name__)

_R = TypeVar("_R")

# Some metaprogramming gotchas:
# The type annotations of a function given to a decorator might not exist
# at the time of the decorator's execution, in the case of "forward references"
# (which is where a type is specified as a str instead of a class/type).
# The return type and the type of `self` can only be checked inside the wrapped
# function and not in the decorator.


def get(link_name: str):
    """
    Creates a decorator for which replaces the given method with one that does a GET request.
    """
    return _http_method_decorator(
        link_name=link_name,
        method_name="GET",
        request=lambda session, url, query: session.get(url, params=query),
    )


def post(link_name: str):
    """
    Creates a decorator for which replaces the given method with one that does a POST request.
    """
    return _http_method_decorator(
        link_name=link_name,
        method_name="POST",
        request=lambda session, url, data: session.post(url, json=data),
    )


def put(link_name: str):
    """
    Creates a decorator for which replaces the given method with one that does a PUT request.
    """
    return _http_method_decorator(
        link_name=link_name,
        method_name="PUT",
        request=lambda session, url, data: session.put(url, json=data),
    )


def delete(link_name: str):
    """
    Creates a decorator for which replaces the given method with one that does a DELETE request.
    """
    return _http_method_decorator(
        link_name=link_name,
        method_name="DELETE",
        request=lambda session, url, _: session.delete(url),
    )


Request = Callable[
    [aiohttp.ClientSession, yarl.URL, dict[str, Any]],
    AsyncContextManager[aiohttp.ClientResponse],
]
"""
A function which does a HTTP request.
"""


def _http_method_decorator(
    link_name: str, method_name: str, request: Request
) -> Callable[
    [Callable[..., Coroutine[None, None, _R]]], Callable[..., Coroutine[None, None, _R]]
]:
    """
    Creates a decorator which transforms a method of a subclass of `aiochris.link.Linked`
    to one which makes an HTTP request.
    """

    def decorator(
        fn: Callable[..., Coroutine[None, None, _R]]
    ) -> Callable[..., Coroutine[None, None, _R]]:
        @functools.wraps(fn)
        async def wrapped(self: Linked, *args, **kwargs) -> _R:
            if args:
                raise TypeError(f"Function {fn} only supports kwargs.")
            return_type = get_return_hint(fn)
            url = self._get_link(link_name)
            data = _filter_none(kwargs)
            logger.debug(f"{method_name} --> {url} : {data}")
            sent = request(self.s, url, data)
            return await deserialize_res(sent, self, data, return_type)

        LinkedMeta.mark_to_check(wrapped, link_name)
        return wrapped

    return decorator


def search(
    collection_name: str, subpath: str = "search/"
) -> Callable[[Callable[..., Search[_R]]], Callable[..., Search[_R]]]:
    """
    Creates a decorator which searches the collection using GET requests.

    (Pagination is handled internally, HTTP requests are made as-needed.)
    """

    def decorator(fn: Callable[..., Search[_R]]) -> Callable[..., Search[_R]]:
        @functools.wraps(fn)
        def wrapped(self: Linked, *args, **kwargs) -> Search[_R]:
            if args:
                raise TypeError(f"Function {fn} only supports kwargs.")
            return_item_type = _get_search_item_type(fn)

            return Search[return_item_type](
                Item=return_item_type,
                client=self,
                base_url=self._get_link(collection_name),
                params=kwargs,
                max_requests=self.max_search_requests,
                subpath=subpath,
            )

        LinkedMeta.mark_to_check(wrapped, collection_name)
        return wrapped

    return decorator


def _get_search_item_type(fn: Callable[..., Search[_R]]) -> Type[_R]:
    return_type = get_return_hint(fn)
    if typing.get_origin(return_type) is not Search:
        raise TypeError(return_type)
    return typing.get_args(return_type)[0]


K = TypeVar("K")
V = TypeVar("V")


def _filter_none(d: dict[K, Optional[V]]) -> dict[K, V]:
    return {k: v for k, v, in d.items() if v is not None}
